from .WBits import WBits

def _format_str_type(x):
    if isinstance(x, str):
       x = str(x).replace('"', '\\"')
       x = "\"" + x + "\""
    return x

class Display(WBits):
    SIZE_SMALL = 0x02
    SIZE_BIG = 0x04
    BUTTON_NONE = 0x01
    BUTTON_L = 0x02
    BUTTON_R = 0x04
    BUTTON_M = 0x08
    def __init__(self, index = 1):
        WBits.__init__(self)
        self.index = index

    
    def register_button(self, cb):
        self._register_event('display{}'.format(self.index), 'button', cb)
    
    def print(self, row, column, text, size = None):
        """
        在某个位置显示内容

        :param row: 显示行数：1~16
        :param column: 显示列数：1~15
        :param text: 显示内容，可以是字符串，整数，小数
        :param size: 设置显示的大小，默认为小号字体  SIZE_SMALL：小号字体，值为2 SIZE_BIG：大号字体（不支持汉字），值为4
        """

        text = _format_str_type(text)
        
        args = []    
        args.append(str(row))
        args.append(str(column))
        args.append(str(text))
        if size != None:
            args.append(str(size))
        command = 'display{}.print({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def draw_dot(self, x, y, page = None):
        """
        在指定坐标画一个点在画点的页使用print函数会导致已经画过的点消失切换到不同的页码在回到画点的页码也会导致已经画过的点消失

        :param x: X轴坐标：1~119
        :param y: Y轴坐标：1~32
        :param page: 显示页数：1~8  默认第1页
        """

        
        args = []    
        args.append(str(x))
        args.append(str(y))
        if page != None:
            args.append(str(page))
        command = 'display{}.draw_dot({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def draw_line(self, head_x, head_y, tail_x, tail_y, page = None):
        """
        通过给定坐标画线段在画线的页使用print函数会导致已经画过的线消失切换到不同的页码在回到画线的页码也会导致已经画过的线消失

        :param head_x: 起始点X轴坐标：1~119
        :param head_y: 起始点Y轴坐标：1~32
        :param tail_x: 终止点X轴坐标：1~119
        :param tail_y: 终止点Y轴坐标：1~32
        :param page: 显示页数：1~8  默认第1页
        """

        
        args = []    
        args.append(str(head_x))
        args.append(str(head_y))
        args.append(str(tail_x))
        args.append(str(tail_y))
        if page != None:
            args.append(str(page))
        command = 'display{}.draw_line({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def draw_chart(self, x, y, page = None):
        """
        画折线图以上次传入的坐标为起点，本次坐标为终点画线段。如果是首次使用，则只画单个点

        :param x: X轴坐标：1~119
        :param y: Y轴坐标：1~32
        :param page: 显示页数：1~8  默认画点在第1页
        """

        
        args = []    
        args.append(str(x))
        args.append(str(y))
        if page != None:
            args.append(str(page))
        command = 'display{}.draw_chart({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def turn_to_page(self, page):
        """
        转到某页

        :param page: 页码：1~8
        """

        
        args = []    
        args.append(str(page))
        command = 'display{}.turn_to_page({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def clear_page(self, page = None):
        """
        清除某页显示的内容

        :param page: 清除的页码：1~8  默认第1页
        """

        
        args = []    
        if page != None:
            args.append(str(page))
        command = 'display{}.clear_page({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def clear_all_pages(self, block = None):
        """
        清除全部8页显示的内容

        :param block: 阻塞参数：  False: 不阻塞 True: 阻塞
        """

        
        args = []    
        if block != None:
            args.append(str(block))
        command = 'display{}.clear_all_pages({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def disable_page_turning(self):
        """
        禁止翻页按键功能禁止翻页按键功能后将不能通过翻页按键来切换不同页码的显示内容系统默认开启翻页按键功能

        """

        command = 'display{}.disable_page_turning()'.format(self.index)
        self._set_command(command)

    
    def enable_page_turning(self):
        """
        开启翻页按键功能系统默认开启翻页按键功能

        """

        command = 'display{}.enable_page_turning()'.format(self.index)
        self._set_command(command)

    
    def get_button_state(self):
        """
        获取翻页按钮状态:rtype: int
        """

        command = 'display{}.get_button_state()'.format(self.index)
        return self._get_command(command)

    
    def set_direction_reverse(self):
        """
        设置显示方向为翻转显示方向，使用该函数后显示内容将会进行180°翻转

        """

        command = 'display{}.set_direction_reverse()'.format(self.index)
        self._set_command(command)

    
    def set_direction_regular(self):
        """
        设置显示方向为系统默认显示方向

        """

        command = 'display{}.set_direction_regular()'.format(self.index)
        self._set_command(command)

    
    def hide_scrollbar(self):
        """
        隐藏页码滚动指示条（屏幕右边的白色小点，用于指示当前页码）系统默认显示页码滚动指示条隐藏后每行最大显示字符数由15变为16

        """

        command = 'display{}.hide_scrollbar()'.format(self.index)
        self._set_command(command)

    
    def show_scrollbar(self):
        """
        显示页码滚动指示条系统默认显示页码滚动指示条

        """

        command = 'display{}.show_scrollbar()'.format(self.index)
        self._set_command(command)

    
    def disable_auto_refresh(self):
        """
        禁止自动刷新显示功能禁止自动刷新后，只能调用刷新函数refresh()才能改变显示内容系统默认开启自动刷新显示功能

        """

        command = 'display{}.disable_auto_refresh()'.format(self.index)
        self._set_command(command)

    
    def enable_auto_refresh(self):
        """
        开启自动刷新显示功能系统默认开启自动刷新显示功能

        """

        command = 'display{}.enable_auto_refresh()'.format(self.index)
        self._set_command(command)

    
    def refresh(self):
        """
        更新一次显示内容在禁止自动刷新显示功能后只能靠此函数来更新显示内容系统默认开启自动刷新显示功能

        """

        command = 'display{}.refresh()'.format(self.index)
        self._set_command(command)

    
    def draw_save_dot(self, x, y, page = None):
        """
        在指定坐标画一个点画点后始终存在，可以使用清屏擦除可与print在同一页显示，显示位置冲突时以画点内容为主

        :param x: X轴坐标：1~119
        :param y: Y轴坐标：1~32
        :param page: 显示页数：1~8  默认第1页
        """

        
        args = []    
        args.append(str(x))
        args.append(str(y))
        if page != None:
            args.append(str(page))
        command = 'display{}.draw_save_dot({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def draw_save_line(self, head_x, head_y, tail_x, tail_y, page = None):
        """
        通过给定坐标画线段画线后始终存在，可以使用清屏擦除可与print在同一页显示，显示位置冲突时以画线内容为主

        :param head_x: 起始点X轴坐标：1~119
        :param head_y: 起始点Y轴坐标：1~32
        :param tail_x: 终止点X轴坐标：1~119
        :param tail_y: 终止点Y轴坐标：1~32
        :param page: 显示页数：1~8  默认第1页
        """

        
        args = []    
        args.append(str(head_x))
        args.append(str(head_y))
        args.append(str(tail_x))
        args.append(str(tail_y))
        if page != None:
            args.append(str(page))
        command = 'display{}.draw_save_line({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def draw_save_chart(self, x, y, page = None):
        """
        画折线图以上次传入的坐标为起点，本次坐标为终点画线段。如果是首次使用，则只画单个点

        :param x: X轴坐标：1~119
        :param y: Y轴坐标：1~32
        :param page: 显示页数：1~8  默认画点在第1页
        """

        
        args = []    
        args.append(str(x))
        args.append(str(y))
        if page != None:
            args.append(str(page))
        command = 'display{}.draw_save_chart({})'.format(self.index, ",".join(args))
        self._set_command(command)

    


from .WBits import WBits

class WBUpload(WBits):
    def __init__(self, index = 1):
        WBits.__init__(self)
        self.index = index

    
    def upload_with_file_path(self, file_path):
        if not self._is_empty(file_path):
            try:
                code = ''
                with open(file_path,'r') as f:
                    code = ''.join(f.readlines())
                if not self._is_empty(code):
                    self._set_raw_command(b'\x05')
                    self._set_command(code)
                    self._set_raw_command(b'\x04')
            except Exception as e:
                print(file_path, '文件不存在！')
                print(e)



    def upload_with_file_content(self, file_content):
        if not self._is_empty(file_content):
            self._set_raw_command(b'\x05')
            self._set_command(file_content)
            self._set_raw_command(b'\x04')
    
    
    def _is_empty(self, arg):
        is_empty = False
        if not arg:
            is_empty = True
            print('参数不可以为空!')
        return is_empty



from .WBUpload import WBUpload
from .MyUtil import MyUtil

class wb_tool(object):
    '''
    豌豆拼工具集合
    '''
    upload = WBUpload()

    @staticmethod
    def show_console():
        '''
        开启控制台输出
        '''
        MyUtil.is_show_console = True

    @staticmethod
    def hide_console():
        '''
        隐藏控制台输出（默认）
        '''
        MyUtil.is_show_console = False


from .MyCore import MyCore
from .MyUtil import MyUtil
import time
import os


class WBits(object):
    '''
    wonderbits parent class
    '''
    # singleton flag
    __init_flag = False

    _wb_serial = None

    def __init__(self):
        if not WBits.__init_flag:
            MyUtil.wb_log('wonderbits初始化')
            WBits.__init_flag = True
            WBits._wb_serial = MyCore()


    def _set_command(self, command):
        '''
        note: send command
        param: command
        return: return 'done' if send command successfully, else return error_msg
        '''
        self._wb_serial.write_command(command)
        self.__timeout_get_command()
        MyUtil.wb_log(MyCore.return_value, '\r\n')
        return MyCore.return_value

    def _get_command(self, command):
        '''
        note get command
        params: command
        return: return value if send command successfully, else return error_msg
        '''
        cmd = 'print({})'.format(command)
        self._wb_serial.write_command(cmd)
        self.__timeout_get_command()
        MyUtil.wb_log(MyCore.return_value, '\r\n')
        return MyCore.return_value

    def __timeout_get_command(self, timeout = 3):
        '''
        max time when execute command,
        if exceed max time, ignore current command.
        '''
        try:
            MyCore.return_value = None
            time_interval = 0.001
            count = timeout // time_interval
            while count > 0:
                if MyCore.return_value:
                    break
                time.sleep(time_interval)
                count = count - 1
                MyCore.can_send_data = True
        except KeyboardInterrupt as e:
            MyUtil.wb_log('exit-wb')
            os._exit(0)


import serial
import serial.tools.list_ports
import threading
import os
import time
from .MyUtil import MyUtil

class MyCore(object):
    '''
    wonderbits mainly handle serial operation（detech, connect, handle_data...）
    '''

    __start_raw_repl_flag = False
    __delete_run_py_flag = False

    # singleton flag
    __init_flag = False

    # get_command_return_value
    return_value = None

    can_send_data = False

    def __init_property(self):
        '''
        init MyCore property
        '''
        self._ser = None
        self._can_used_serial_port = []

    def __init__(self):
        if not MyCore.__init_flag:
            MyCore.__init_flag = True
            self.__init_property()
            threading.Thread(target=self._try_connect_serial_thread, args=('connect_serial_thread',)).start()

    def _try_connect_serial_thread(self, thread_name):
        '''
        try to connect serial
        '''
        try:
            while True:
                self._init_connect()
                can_used_serial_count = len(self._can_used_serial_port)
                if not can_used_serial_count:
                    MyUtil.wb_log('无可用串口')
                    os._exit(0)
                else:
                    if self._ser == None and can_used_serial_count != 0:
                        self._connect_serial()
                    time.sleep(.5)
        except KeyboardInterrupt as e:
            MyUtil.wb_log('退出豌豆拼')
            os._exit(0)
        except OSError as e:
            MyUtil.wb_log('连接异常')
            os._exit(0)
        

    def _init_connect(self):
        '''
        get serial ports info
        '''
        if self._ser != None:
            return
        self.__init_property()
        port_list = list(serial.tools.list_ports.comports())
        for i in range(len(port_list)):
            port = port_list[i]
            if port.pid == 29987 or port.pid == 60000:
                self._can_used_serial_port.append(port)
                MyUtil.wb_log(port, port.pid)

    def _connect_serial(self):
        '''
        connect serial
        '''
        try:
            portx = self._can_used_serial_port[0].device
            bps = 115200
            timex = 1
            self._ser = serial.Serial(portx, bps, timeout=timex)
            cmd = MyUtil.wb_encode('{}\r\n'.format("reset()"))
            self._ser.write(cmd)
            # self._start_raw_repl()
            threading.Thread(target=self._prepare_communication, args=('handle_serial_port_target',)).start()
        except serial.serialutil.SerialException as e:
            MyUtil.wb_log('串口异常{}',format(e))
            self.__init_property()
        

    def _prepare_communication(self, target_name):
        '''
        brefore communication
        do enter raw repl
        and delete run_loop.py
        and soft reboot
        '''
        try:
            buffer = ''
            while True:
                oneByte = self._ser.read(1)
                try:
                    oneChar = MyUtil.wb_decode(oneByte)
                except:
                    MyUtil.wb_log('解析数据失败 {}'.format(oneByte))
                    continue
                if oneChar:
                    buffer += oneChar
                    if oneByte == b'\n':
                        MyUtil.wb_log(buffer)
                        if 'raw REPL; CTRL-B to exit\r\n' == buffer:
                            # second step: delete run py
                            if not self.__delete_run_py_flag:
                                self._delete_run_py_repl()
                            else:
                                MyUtil.wb_log('已成功切换到raw repl mode: 可以正常通信了!', '\r\n')
                                MyCore.can_send_data = True
                                threading.Thread(target=self._normal_communication, args=('normal_communication_thread',)).start()
                                break
                        if buffer.startswith('>OK'):
                            # third step(last step): soft reboot enter raw repl
                            self._ser.write(b'\x04')
                        buffer = ''
                    if oneChar == '>' and buffer[-3:] == '>>>':
                        MyUtil.wb_log(buffer)
                        buffer = ''
                        # first step: enter raw repl mode
                        self._start_raw_repl()
        except OSError as e:
            MyUtil.wb_log('连接异常')
            os._exit(0)

    def _start_raw_repl(self):
        '''
        start enter raw repl mode
        '''
        if not self.__start_raw_repl_flag:
            MyUtil.wb_log('开始进入raw repl mode')
            self.__start_raw_repl_flag = True
            first_command_list = [b'\n\x03\x03', b'\n\x01']
            for command in first_command_list:
                self._ser.write(command)
    
    def _delete_run_py_repl(self):
        '''
        delete run_loop.py
        '''
        if not self.__delete_run_py_flag:
            MyUtil.wb_log('开始删除run_loop.py')
            self.__delete_run_py_flag = True
            # note: empty_char must be (leng=1) char;
            empty_char = ' '
            delete_run_py_command = "try:\r\n" + empty_char * 4 + "import os\r\nexcept ImportError:\r\n" + empty_char * 4 + "import uos as os\r\nos.remove('run_loop.py')\r\n"
            delete_run_py_end_command = b'\x04'
            self._ser.write(MyUtil.wb_encode(delete_run_py_command))
            self._ser.write(delete_run_py_end_command)

    def _normal_communication(self, target_name):
        '''
        handle communication data
        '''
        try:
            byte_buffer = b''
            buffer = ''
            while True:
                oneByte = self._ser.read(1)
                try:
                    oneChar = MyUtil.wb_decode(oneByte)
                    if oneChar:
                        buffer += oneChar
                except:
                    MyUtil.wb_log('解析数据失败 {}'.format(oneByte))
                    continue
                if oneByte:
                    byte_buffer += oneByte
                    if (byte_buffer.startswith(b'OK') or byte_buffer.startswith(b'>OK')) and byte_buffer.endswith(b'\x04>'):
                        MyUtil.wb_log(byte_buffer)
                        # parse get_command_return_value
                        get_command_return_value = MyUtil.parse_data_from_raw_repl(buffer)
                        MyCore.return_value = get_command_return_value
                        byte_buffer = b''
                        buffer = ''
                        MyCore.can_send_data = True
        except OSError as e:
            MyUtil.wb_log('连接异常')
            os._exit(0)

    def write_command(self, command):
        try:
            cmd = MyUtil.wb_encode(command) + b'\x04'
            while not MyCore.can_send_data:
                time.sleep(.001)
            if MyCore.can_send_data:
                MyUtil.wb_log(cmd)
                self._ser.write(cmd)
                MyCore.can_send_data = False
        except KeyboardInterrupt as e:
            MyUtil.wb_log('exit-wb')
            os._exit(0)



# if __name__ == "__main__":

#     wbits = WBits()
#     for i in range(10):
#         wbits._send_command('display1.print(1,1,{})'.format(i))
#         value = wbits._get_command('control1.get_sw4()')




from .WBits import WBits

def _format_str_type(x):
    if isinstance(x, str):
       x = str(x).replace('"', '\\"')
       x = "\"" + x + "\""
    return x

class Control(WBits):
    def __init__(self, index = 1):
        WBits.__init__(self)
        self.index = index

    
    def register_sw1(self, cb):
        self._register_event('control{}'.format(self.index), 'sw1', cb)
    
    def register_sw2(self, cb):
        self._register_event('control{}'.format(self.index), 'sw2', cb)
    
    def register_sw3(self, cb):
        self._register_event('control{}'.format(self.index), 'sw3', cb)
    
    def register_sw4(self, cb):
        self._register_event('control{}'.format(self.index), 'sw4', cb)
    
    def register_m1(self, cb):
        self._register_event('control{}'.format(self.index), 'm1', cb)
    
    def register_m2(self, cb):
        self._register_event('control{}'.format(self.index), 'm2', cb)
    
    def register_m1_value(self, cb):
        self._register_event('control{}'.format(self.index), 'm1_value', cb)
    
    def register_m2_value(self, cb):
        self._register_event('control{}'.format(self.index), 'm2_value', cb)
    
    def is_sw1_pressed(self):
        """
        判断按键SW1是否被按下:rtype: bool
        """

        command = 'control{}.is_sw1_pressed()'.format(self.index)
        return self._get_command(command)

    
    def is_sw2_pressed(self):
        """
        判断按键SW2是否被按下:rtype: bool
        """

        command = 'control{}.is_sw2_pressed()'.format(self.index)
        return self._get_command(command)

    
    def is_sw3_at_1(self):
        """
        判断SW3的是否在‘1’的位置（‘1’指的是电路上白色的数字）:rtype: bool
        """

        command = 'control{}.is_sw3_at_1()'.format(self.index)
        return self._get_command(command)

    
    def get_sw4(self):
        """
        获取SW4的位置值:rtype: int
        """

        command = 'control{}.get_sw4()'.format(self.index)
        return self._get_command(command)

    
    def is_m1_connected(self):
        """
        判断获取M1与COM是否导通一般的使用方法是：将连接线插入到控制模块的接头上，实验者一手握住COM线头（黑色），另一手握住M1或M2线头（黄或绿色）。导通时板子上相应指示灯会亮起:rtype: bool
        """

        command = 'control{}.is_m1_connected()'.format(self.index)
        return self._get_command(command)

    
    def is_m2_connected(self):
        """
        判断获取M2与COM是否导通一般的使用方法是：将连接线插入到控制模块的接头上，实验者一手握住COM线头（黑色），另一手握住M1或M2线头（黄或绿色）。导通时板子上相应指示灯会亮起:rtype: bool
        """

        command = 'control{}.is_m2_connected()'.format(self.index)
        return self._get_command(command)

    
    def set_m1_m2_sensitivity(self, limit):
        """
        设置M1和M2灵敏度灵敏度越高，is_m1_connected()和is_m2_connected()越容易返回True

        :param limit: 灵敏度：0~100
        """

        
        args = []    
        args.append(str(limit))
        command = 'control{}.set_m1_m2_sensitivity({})'.format(self.index, ",".join(args))
        self._set_command(command)

    
    def get_m1_value(self):
        """
        获取M1的电阻率:rtype: float
        """

        command = 'control{}.get_m1_value()'.format(self.index)
        return self._get_command(command)

    
    def get_m2_value(self):
        """
        获取M2的电阻率:rtype: float
        """

        command = 'control{}.get_m2_value()'.format(self.index)
        return self._get_command(command)

    



class MyUtil(object):
    '''
    wonderbits util set
    '''
    # decide if show console
    is_show_console = False

    @staticmethod
    def wb_log(*params):
        '''
        接受不定个数的参数, 并print出来
        '''
        if not MyUtil.is_show_console:
            return
        print('###wb-log: ', end="")
        for param in params:
            print(param, end=" ")
        print()

    @staticmethod
    def wb_decode(raw_byte = b''):
        return raw_byte.decode('gbk')

    @staticmethod
    def wb_encode(raw_str = ''):
        return raw_str.encode('gbk')

    @staticmethod
    def parse_data_from_raw_repl(buffer):
        OK = 'OK'
        beginIndex = buffer.find(OK) + len(OK)
        endIndex = buffer.rfind('\r\n')
        # set_command finished successfully
        if endIndex == -1:
            return 'done'
        # set_command failed
        # get_command finished successfully
        # get_command failed
        return buffer[beginIndex: endIndex]